//
//  DNParser.m
//  dNodi
//
//  Created by Derek Clarkson on 7/22/11.
//  Copyright 2011. All rights reserved.
//

#import <dUsefulStuff/DCCommon.h>
#import <dUsefulStuff/NSObject+dUsefulStuff.h>

#import <dNodi/DNParser.h>
#import <dNodi/DNSelectorFactory.h>
#import <dNodi/DNCommon.h>
#import <dNodi/DNScanner.h>

@interface DNParser () {
@private
	NSDictionary *_syntax;
	DNToken _lastToken;
}

@property (nonatomic, retain) id<DNSelector> firstSelector;
@property (nonatomic, retain) id<DNSelector> lastSelector;

-(BOOL) syntaxIsValidWithToken:(DNToken) toToken
						  atLocation:(NSUInteger) location
								 error:(NSError **) error;
@end

@implementation DNParser

@synthesize selectorFactory = _selectorFactory;
@synthesize firstSelector = _firstSelector;
@synthesize lastSelector = _lastSelector;

#pragma mark - Lifecycle

- (void) dealloc {
	self.selectorFactory = nil;
	self.firstSelector = nil;
	self.lastSelector = nil;
	DC_DEALLOC(_syntax);
	[super dealloc];
}

-(id) init {
	self = [super init];
	if (self) {
		_syntax = [@{
					  @(DNTokenStart) : @[ @(DNTokenNodeName), @(DNTokenSlash), @(DNTokenOpenBracket), @(DNTokenSubtree), @(DNTokenParent) ],
					  @(DNTokenSlash) : @[ @(DNTokenNodeName), @(DNTokenOpenBracket), @(DNTokenParent), @(DNTokenEnd) ],
					  @(DNTokenNodeName) : @[@(DNTokenSlash), @(DNTokenOpenBracket), @(DNTokenSubtree), @(DNTokenParent),@(DNTokenEnd) ],
					  @(DNTokenOpenBracket) : @[@(DNTokenNumber), @(DNTokenAttributeName) ],
					  @(DNTokenAttributeName) : @[@(DNTokenProperty), @(DNTokenEquals) ],
					  @(DNTokenEquals) : @[@(DNTokenOpenQuote) ],
					  @(DNTokenOpenQuote) : @[@(DNTokenValue) ],
					  @(DNTokenValue) : @[@(DNTokenCloseQuote) ],
					  @(DNTokenCloseQuote) : @[@(DNTokenCloseBracket), @(DNTokenComma) ],
					  @(DNTokenComma) : @[@(DNTokenAttributeName) ],
					  @(DNTokenCloseBracket) : @[@(DNTokenSlash), @(DNTokenEnd), @(DNTokenOpenBracket), @(DNTokenParent) ],
					  @(DNTokenNumber) : @[@(DNTokenCloseBracket) ],
					  @(DNTokenSubtree) : @[@(DNTokenNodeName), @(DNTokenOpenBracket) ],
					  @(DNTokenParent) : @[@(DNTokenSlash), @(DNTokenEnd) ],
					  @(DNTokenProperty) : @[@(DNTokenAttributeName) ]
					  } retain];
		_lastToken = DNTokenStart;
		self.selectorFactory = [[DNSelectorFactory alloc] initWithSelectorFactoryDelegate:self];
	}
	return self;
}

-(id<DNSelector>) selectorChainForQuery:(NSString *) query error:(NSError **) error {
	
	DNScanner *scanner = [[DNScanner alloc] initWithDelegate:self];
	BOOL successful = [scanner scanPath:query error:error];
	if (!successful) {
		return nil;
	}
	
	// Return the first selector.
	return self.firstSelector;
	
}

#pragma mark - Delegates

-(BOOL) didScanToken:(DNToken) token
			 atLocation:(NSUInteger) location
			  withValue:(id) value
					error:(NSError **) error {
	return token != DNTokenError
	&& [self syntaxIsValidWithToken:token atLocation:location error:error]
	&& [_selectorFactory tokenRead:token value:value error:error];
}

-(BOOL) scanDidEndWithError:(NSError **) error {
	return [self syntaxIsValidWithToken:DNTokenEnd atLocation:0 error:error];
}

-(void) didCreateSelector:(id<DNSelector>) selector {
	
	// If the first selector is not set.
	if (self.firstSelector == nil) {
		self.firstSelector = selector;
		self.lastSelector = selector;
		return;
	}
	
	// Chain the selector.
	self.lastSelector.nextSelector = selector;
	self.lastSelector = selector;
	
}

#pragma mark - Private

-(BOOL) syntaxIsValidWithToken:(DNToken) toToken
						  atLocation:(NSUInteger) location
								 error:(NSError **) error {
	NSArray * validToTokens = _syntax[@(_lastToken)];
	if (validToTokens == nil || ![validToTokens containsObject:@(toToken)]) {
		DC_LOG(@"Invalid query, returning error at scan location: %i", location);
		
		if (toToken == DNTokenEnd) {
			[self setError:error
						 code:DNErrorInvalidNodiExpression
				errorDomain:DN_ERROR_DOMAIN
		 shortDescription:@"Incomplete query expression"
			 failureReason:@"Query appears to be incomplete."];
		} else {
			[self setError:error
						 code:DNErrorInvalidNodiExpression
				errorDomain:DN_ERROR_DOMAIN
		 shortDescription:@"Invalid Query"
			 failureReason:[NSString stringWithFormat:@"Cannot go from a %@ to a %@ at character index %i", tokenDescriptions[_lastToken], tokenDescriptions[toToken], location]];
		}
		
		return NO;
	}
	_lastToken = toToken;
	return YES;
}

@end
