//
//  DNScanner.m
//  dNodi
//
//  Created by Derek Clarkson on 23/11/12.
//
//

#import <dUsefulStuff/DCCommon.h>
#import <dUsefulStuff/NSObject+dUsefulStuff.h>
#import <dNodi/DNScanner.h>
#import <dNodi/DNCommon.h>

@interface DNScanner () {
@private
	
	id<DNScannerDelegate> _delegate;
	
	NSScanner *_scanner;
	NSMutableCharacterSet *_validNodeNameCharacters;
	NSMutableString *_attributeName;
	NSString *_quote;
	
	BOOL _scanningAttributes;
	BOOL _scanningValue;
}
- (DNToken) scanNextTokenAtCharacterIndex:(NSUInteger) scanIndex
									 intoNodeValue:(id *)nodeValue
										  lastToken:(DNToken) lastToken
												error:(NSError **) error;
@end

@implementation DNScanner

-(void) dealloc {
	DC_DEALLOC(_validNodeNameCharacters);
	DC_DEALLOC(_scanner);
	DC_DEALLOC(_attributeName);
	DC_DEALLOC(_quote);
	[super dealloc];
}

-(id) initWithDelegate:(id<DNScannerDelegate>) delegate {
	self = [super init];
	if (self) {
		_delegate = delegate;
		// Node names can be any Alpha numeric plus underscores.
		_validNodeNameCharacters = [[NSCharacterSet alphanumericCharacterSet] mutableCopy];
		[_validNodeNameCharacters addCharactersInString:@"_*"];
	}
	return self;
}

-(BOOL) scanPath:(NSString *)path error:(NSError **)error {
	
	DC_LOG(@"Scanning '%@'", path);
	
	// Create a new instance of NSScanner.
	_scanner = [[NSScanner alloc] initWithString:path];
	_scanningAttributes = NO;
	
	// Process.
	DNToken token = DNTokenStart;
	NSUInteger scanningFrom = 0;
	while (![_scanner isAtEnd]) {
		
		// Track the scanning start location for error messages.
		scanningFrom = [_scanner scanLocation];
		
      // Get the next token and it's value.
		id nodeValue = nil;
		token = [self scanNextTokenAtCharacterIndex:scanningFrom
												intoNodeValue:&nodeValue
													 lastToken:token
														  error:error];
		if (token == DNTokenError) {
			return NO;
		}
		
      // Notify the delegate.
		if (![_delegate didScanToken:token
								atLocation:scanningFrom
								 withValue:nodeValue
									  error:error]) {
			return NO;
		}
		
	}
	
	// Finally tell the delegate we have finished by sending a end token.
	return [_delegate scanDidEndWithError:error];

}

- (DNToken) scanNextTokenAtCharacterIndex:(NSUInteger) scanIndex
									 intoNodeValue:(id *)nodeValue
										  lastToken:(DNToken) lastToken
												error:(NSError **) error {
	
   // Temp vars to hold data read by the scanner.
	int value = 0;
	
	// first we check if we are expecting a free form string value.
	if (_scanningValue) {
		[_scanner scanUpToString:_quote intoString:nodeValue];
      DC_LOG(@"Found value '%@'", *nodeValue);
		_scanningValue = NO;
		return DNTokenValue;
	}

	if ([_scanner scanString:@"/.." intoString:NULL]) {
		DC_LOG(@"Found parent reference (/..)");
		return DNTokenParent;
	}

	// Now we do the more format parsing.
	if ([_scanner scanString:@"//" intoString:NULL]) {
		DC_LOG(@"Found anywhere indicator (//)");
		return DNTokenSubtree;
	}
	
	if ([_scanner scanString:@"/" intoString:NULL]) {
		DC_LOG(@"Found%@ slash (/)", scanIndex == 0 ? @" root" : @"");
		return DNTokenSlash;
	}
	
	if ([_scanner scanString:@"[" intoString:NULL]) {
		DC_LOG(@"Found start bracket ([)");
		_scanningAttributes = YES;
		return DNTokenOpenBracket;
	}
	
	if ([_scanner scanString:@"]" intoString:NULL]) {
		DC_LOG(@"Found closing bracket (])");
		_scanningAttributes = NO;
		return DNTokenCloseBracket;
	}
		
	if ([_scanner scanString:@"," intoString:NULL]) {
		DC_LOG(@"Found comma (,)");
		return DNTokenComma;
	}
	
	if ([_scanner scanString:@"." intoString:NULL]) {
		DC_LOG(@"Found dot notation (.)");
		return DNTokenProperty;
	}
	
	if ([_scanner scanString:@"'" intoString:nodeValue] || [_scanner scanString:@"\"" intoString:nodeValue]) {
		if (lastToken == DNTokenValue) {
			DC_LOG(@"Found closing quote %@", _quote);
			DC_DEALLOC(_quote);
			return DNTokenCloseQuote;
		} else {
			// Starting a new quoted value.
			_quote = [*nodeValue retain];
			DC_LOG(@"Found opening quote %@", _quote);
			_scanningValue = YES;
			return DNTokenOpenQuote;
		}
	}
	
	if ([_scanner scanString:@"=" intoString:NULL]) {
		DC_LOG(@"Found =");
		return DNTokenEquals;
	}
	
	if ([_scanner scanInt:&value]) {
		DC_LOG(@"Found arrray index %i", value);
		*nodeValue = @(value);
		return DNTokenNumber;
   }
	
	if ([_scanner scanCharactersFromSet:_validNodeNameCharacters intoString:nodeValue]) {
		DC_LOG(@"Found %@ name \"%@\"",(_scanningAttributes ? @"attribute" : @"node"), *nodeValue);
		if (_scanningAttributes) {
			if ([*nodeValue rangeOfString:@"*"].location != NSNotFound) {
				[self setError:error
							 code:DNErrorInvalidNodiExpression
					errorDomain:DN_ERROR_DOMAIN
			 shortDescription:@"Invalid attribute name"
				 failureReason:[NSString stringWithFormat:@"Attribute name '%@'at position %i", *nodeValue, scanIndex]];
				return DNTokenError;
			}
			return DNTokenAttributeName;
		} else {
			return DNTokenNodeName;
		}
	}
	
	// Nothing scanned so throw an error because we don't know what it is.
	DC_LOG(@"Unknown token in query, returning error");
	[self setError:error
				 code:DNErrorInvalidNodiExpression
		errorDomain:DN_ERROR_DOMAIN
 shortDescription:@"Unknown token"
	 failureReason:[NSString stringWithFormat:@"Unknown syntax encountered in query at position %i after %@", scanIndex, tokenDescriptions[lastToken]]];
	return DNTokenError;
}


@end
