//
//  DNParserTests.m
//  dNodi
//
//  Created by Derek Clarkson on 7/24/11.
//  Copyright 2011. All rights reserved.
//

#import <GHUnitIOS/GHUnit.h>
#import <dUsefulStuff/DCCommon.h>

#import <dNodi/DNParser.h>
#import <dNodi/DNSelectorFactory.h>
#import <dNodi/DNScanner.h>
#import <dNodi/DNSelector.h>

#import <dNodi/DNNodeNameSelector.h>
#import <dNodi/DNIndexedSelector.h>
#import <dNodi/DNAttributeSelector.h>
#import <dNodi/DNParentSelector.h>
#import <dNodi/DNChildNodesSelector.h>
#import <dNodi/DNTreeSearchSelector.h>

@interface DNParserTests : GHTestCase {
@private
	DNParser *_parser;
	NSError *_error;
}
-(id<DNSelector>) successfulPath:(NSString *) path returnsSelectorClasses:(NSArray *) selectorClasses;
-(void) unSuccessfulPath:(NSString *) path withMessage:(NSString *) message;
@end

@implementation DNParserTests

-(void) setUp {
	_parser = [[DNParser alloc] init];
	_error = NULL;
}

-(void) tearDown {
	DC_DEALLOC(_parser);
}

-(void) testStartingWithANodeName {
	[self successfulPath:@"abc" returnsSelectorClasses:@[[DNNodeNameSelector class]]];
}

-(void) testStartingWithAWildcardNodeName {
	[self successfulPath:@"abc*" returnsSelectorClasses:@[[DNNodeNameSelector class]]];
}

-(void) testStartingWithSubnodes {
	[self successfulPath:@"/" returnsSelectorClasses:@[[DNChildNodesSelector class]]];
}

-(void) testStartingWithIndex {
	[self successfulPath:@"[0]" returnsSelectorClasses:@[[DNIndexedSelector class]]];
}

-(void) testStartingWithParent {
	[self successfulPath:@"/.." returnsSelectorClasses:@[[DNParentSelector class]]];
}

-(void) testStartingWithAttributes {
	[self successfulPath:@"[abc='def']" returnsSelectorClasses:@[[DNAttributeSelector class]]];
}

-(void) testSearchAndNodeName {
	[self successfulPath:@"//abc" returnsSelectorClasses:@[[DNTreeSearchSelector class], [DNNodeNameSelector class]]];
}

-(void) testSubnodesThenNodeName {
	[self successfulPath:@"/abc" returnsSelectorClasses:@[[DNChildNodesSelector class], [DNNodeNameSelector class]]];
}

-(void) testSubnodesThenNodeNameAndIndex {
	[self successfulPath:@"/abc[2]" returnsSelectorClasses:@[[DNChildNodesSelector class], [DNNodeNameSelector class], [DNIndexedSelector class]]];
}

-(void) testNodeNameAndAttribute {
	[self successfulPath:@"abc[abc='def']" returnsSelectorClasses:@[[DNNodeNameSelector class], [DNAttributeSelector class]]];
}

-(void) testNodeNameAttributesAndIndexed {
	[self successfulPath:@"/abc[abc='def'][1]" returnsSelectorClasses:@[[DNChildNodesSelector class], [DNNodeNameSelector class], [DNAttributeSelector class], [DNIndexedSelector class]]];
}

-(void) testMultipleAttributes {
	[self successfulPath:@"/abc[abc='def',ghi='mno']" returnsSelectorClasses:@[[DNChildNodesSelector class], [DNNodeNameSelector class], [DNAttributeSelector class], [DNAttributeSelector class]]];
}

-(void) testAttributeWithProperty {
	[self successfulPath:@"/abc[abc.ghi='def']" returnsSelectorClasses:@[[DNChildNodesSelector class], [DNNodeNameSelector class], [DNAttributeSelector class]]];
}

-(void) testTreeSearchThenProperty {
	[self successfulPath:@"//[abc='def']" returnsSelectorClasses:@[[DNTreeSearchSelector class], [DNAttributeSelector class]]];
}

-(void) testErrorAttributesWithoutBrackets {
	[self unSuccessfulPath:@"abc='def'" withMessage:@"Cannot go from a Node name to a '=' at character index 3"];
}

-(void) testErrorAttributesMissingStartingQuote {
	[self unSuccessfulPath:@"[abc=def']" withMessage:@"Cannot go from a '=' to a Attribute name at character index 5"];
}

-(void) testErrorAttributesMissingEndingQuote {
	[self unSuccessfulPath:@"[abc='def]" withMessage:@"Query appears to be incomplete."];
}

-(void) testErrorAttributesValueOutsideQuote {
	[self unSuccessfulPath:@"[abc='def'ghi]" withMessage:@"Cannot go from a Closing ' or \" to a Attribute name at character index 10"];
}

-(void) testErrorInvalidIndex {
	[self unSuccessfulPath:@"[!]" withMessage:@"Unknown syntax encountered in query at position 1 after '['"];
}

-(void) testErrorAttributesExtraComma {
	[self unSuccessfulPath:@"[abc='ghi',]" withMessage:@"Cannot go from a ',' to a ']' at character index 11"];
}

-(void) testErrorAttributeTrailingPropertyDot {
	[self unSuccessfulPath:@"[abc.='ghi']" withMessage:@"Cannot go from a '.' to a '=' at character index 5"];
}

#pragma mark - Helpers

-(id<DNSelector>) successfulPath:(NSString *) path returnsSelectorClasses:(NSArray *) selectorClasses {
	
	id<DNSelector> selector = [_parser selectorChainForQuery:path error:&_error];
	GHAssertNotNil(selector, @"Error: %@", [_error localizedFailureReason]);
	GHAssertTrue(_error == NULL, nil);

	// CHeck each selector against the expected classes and that there are the same number of selectors as classes.
	NSUInteger idx = 0;
	while (selector != nil && idx < [selectorClasses count]) {
		GHAssertTrue([selector isKindOfClass:selectorClasses[idx]], @"Got a %@, Expecting a %@", NSStringFromClass(selectorClasses[idx]), NSStringFromClass([selector class]));
		selector = selector.nextSelector;
		idx++;
	}
	GHAssertTrue(idx == [selectorClasses count], @"Selector counts don't match.");
	return selector;
}

-(void) unSuccessfulPath:(NSString *) path withMessage:(NSString *) message {
	id<DNSelector> selector = [_parser selectorChainForQuery:path error:&_error];
	GHAssertNil(selector, nil);
	GHAssertTrue(_error != NULL, nil);
	GHAssertEqualStrings([_error localizedFailureReason], message, nil);
}

@end
