//
//  DNScanner.m
//  dNodi
//
//  Created by Derek Clarkson on 23/11/12.
//  Copyright (c) 2012. All rights reserved.
//

#import <GHUnitIOS/GHUnit.h>
#import <dUsefulStuff/DCCommon.h>
#import <dNodi/DNScanner.h>
#import <dNodi/DNScannerDelegate.h>

@interface DNScannerTests : GHTestCase<DNScannerDelegate> {
@private
	NSMutableArray *_tokens;
	NSMutableArray *_values;
	BOOL _scanEndedSuccessfully;
	NSError *_error;
	DNScanner *_scanner;
}
-(void) expectSuccessfulScan:(NSString *) path expectedTokens:(NSArray *) expectedTokens expectedValues:(NSArray *) expectedValues;
-(void) expectFailedScan:(NSString *) path;
@end

@implementation DNScannerTests

-(void) setUp {
	_tokens = [[NSMutableArray alloc] init];
	_values = [[NSMutableArray alloc] init];
	_scanEndedSuccessfully = NO;
	_error = NULL;
	_scanner = [[DNScanner alloc] initWithDelegate:self];
}

-(void) tearDown {
	DC_DEALLOC(_tokens);
	DC_DEALLOC(_values);
	DC_DEALLOC(_scanner);
}

-(void) testScanAttributeValue {
	NSArray *tokens = @[@(DNTokenOpenBracket),
							@(DNTokenAttributeName),
							@(DNTokenEquals),
							@(DNTokenOpenQuote),
							@(DNTokenValue),
							@(DNTokenCloseQuote),
							@(DNTokenCloseBracket)];
	NSArray *values = @[[NSNull null],
							@"abc",
							[NSNull null],
							@"'",
							@"def",
							@"'",
							[NSNull null]];
	[self expectSuccessfulScan:@"[abc='def']" expectedTokens:tokens expectedValues:values];
}

-(void) testScanPropertyAttributeValue {
	NSArray *tokens = @[@(DNTokenOpenBracket),
							@(DNTokenAttributeName),
							@(DNTokenProperty),
							@(DNTokenAttributeName),
							@(DNTokenEquals),
							@(DNTokenOpenQuote),
							@(DNTokenValue),
							@(DNTokenCloseQuote),
							@(DNTokenCloseBracket)];
	NSArray *values = @[[NSNull null],
							@"abc",
							[NSNull null],
							@"ghi",
							[NSNull null],
							@"'",
							@"def",
							@"'",
							[NSNull null]];
	[self expectSuccessfulScan:@"[abc.ghi='def']" expectedTokens:tokens expectedValues:values];
}

-(void) testScanMultipleAttributes {
	NSArray *tokens = @[@(DNTokenOpenBracket),
							@(DNTokenAttributeName),
							@(DNTokenEquals),
							@(DNTokenOpenQuote),
							@(DNTokenValue),
							@(DNTokenCloseQuote),
							@(DNTokenComma),
							@(DNTokenAttributeName),
							@(DNTokenEquals),
							@(DNTokenOpenQuote),
							@(DNTokenValue),
							@(DNTokenCloseQuote),
							@(DNTokenCloseBracket)];
	NSArray *values = @[[NSNull null],
							@"abc",
							[NSNull null],
							@"'",
							@"def",
							@"'",
							[NSNull null],
							@"ghi",
							[NSNull null],
							@"'",
							@"jkl",
							@"'",
							[NSNull null]];
	[self expectSuccessfulScan:@"[abc='def',ghi='jkl']" expectedTokens:tokens expectedValues:values];
}

-(void) testScanNodeName {
	[self expectSuccessfulScan:@"abc" expectedTokens: @[@(DNTokenNodeName)] expectedValues:@[@"abc"]];
}

-(void) testScanWildcardSuffixNodeName {
	[self expectSuccessfulScan:@"abc*" expectedTokens: @[@(DNTokenNodeName)] expectedValues:@[@"abc*"]];
}

-(void) testScanWildcardPrefixNodeName {
	[self expectSuccessfulScan:@"*abc" expectedTokens: @[@(DNTokenNodeName)] expectedValues:@[@"*abc"]];
}

-(void) testScanWildcardNodeName {
	[self expectSuccessfulScan:@"ab*c" expectedTokens: @[@(DNTokenNodeName)] expectedValues:@[@"ab*c"]];
}

-(void) testScanMultipleWildcardNodeName {
	[self expectSuccessfulScan:@"ab*c*" expectedTokens: @[@(DNTokenNodeName)] expectedValues:@[@"ab*c*"]];
}

-(void) testScanSubnodes {
	[self expectSuccessfulScan:@"//" expectedTokens: @[@(DNTokenSubtree)] expectedValues:@[[NSNull null]]];
}

-(void) testScanParent {
	[self expectSuccessfulScan:@"/.." expectedTokens: @[@(DNTokenParent)] expectedValues:@[[NSNull null]]];
}

-(void) testScanChildnodes {
	NSArray *tokens = @[@(DNTokenNodeName),
							@(DNTokenSlash)];
	NSArray *values = @[@"abc",
							[NSNull null]];
	[self expectSuccessfulScan:@"abc/" expectedTokens:tokens expectedValues:values];
}

-(void) testFailsWhenInvalidNodeNames {
	[self expectFailedScan:@"@#$%"];
}

-(void) testFailsWhenWildcardUsedInAttributeName {
	[self expectFailedScan:@"[abc*='def']"];
}

#pragma mark - Helpers

-(void) expectSuccessfulScan:(NSString *) path expectedTokens:(NSArray *) expectedTokens expectedValues:(NSArray *) expectedValues {
	[_scanner scanPath:path error:&_error];
	GHAssertTrue(_scanEndedSuccessfully, nil);
	GHAssertTrue(_error == NULL, nil);
	GHAssertEquals([_tokens count], [expectedTokens count], nil);
	GHAssertEqualObjects(_tokens, expectedTokens, nil);
	GHAssertEqualObjects(_values, expectedValues, nil);
}

-(void) expectFailedScan:(NSString *) path  {
	[_scanner scanPath:path error:&_error];
	GHAssertFalse(_scanEndedSuccessfully, nil);
	GHAssertTrue(_error != NULL, nil);
}

#pragma mark - Delegates methods

-(BOOL) didScanToken:(DNToken) token
			 atLocation:(NSUInteger) location
			  withValue:(id) value
					error:(NSError **) error {
	[_tokens addObject:@(token)];
	[_values addObject:value == nil ? [NSNull null] : value];
	return YES;
}

-(BOOL) scanDidEndWithError:(NSError **) error {
	_scanEndedSuccessfully = YES;
	return YES;
}


@end
